/*
 * 				Shell Library Functions
 *
 * Copyright 2005 Johannes Anderwald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"
#include "wine/port.h"

#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include "winerror.h"
#include "windef.h"
#include "winbase.h"
#include "winreg.h"
#include "wingdi.h"
#include "winuser.h"
#include "commdlg.h"
#include "wine/debug.h"

#include "shellapi.h"
#include <shlwapi.h>
#include "shlobj.h"
#include "shell32_main.h"
#include "shresdef.h"
#include "undocshell.h"
#include <prsht.h>


#define MAX_PROPERTY_SHEET_PAGE	32

  typedef struct _LANGANDCODEPAGE_ 
  {
    WORD lang;
    WORD code;
} LANGANDCODEPAGE, *LPLANGANDCODEPAGE;




/*************************************************************************
 *
 * SH_InsertPropertySheetPage [Internal]
 *
 * stores a HPROPSHEETPAGE handle in an array
 *
 */

BOOL
SH_InsertPropertySheetPage(HPROPSHEETPAGE * hppages, HPROPSHEETPAGE hpage, int numppages)
{
 int index;
 int freeindex = -1;

 if(hppages == NULL)
   return FALSE;

 for(index = 0; index < numppages; index++)
 {
   if(hppages[index] == NULL)
   {
	 if(freeindex == -1)
		 freeindex = index;
   }
   if(hppages[index] == hpage)
     return TRUE;
 }
 if(freeindex < 0)
   return FALSE;

 hppages[freeindex] = hpage;
 return TRUE;
}

/*************************************************************************
 *
 * SH_CreatePropertySheetPage [Internal]
 *
 * creates a property sheet page from an resource name
 *
 */


HPROPSHEETPAGE
SH_CreatePropertySheetPage(LPSTR resname, DLGPROC dlgproc, LPFNPSPCALLBACKW pfnCallback, LPARAM lParam)
{
  HRSRC hRes;
  LPVOID lpsztemplate;
  PROPSHEETPAGEW ppage;

  if(resname == NULL)
   return (HPROPSHEETPAGE)0;

  hRes = FindResourceA(shell32_hInstance, resname, (LPSTR)RT_DIALOG);
  if(hRes == NULL)
    return (HPROPSHEETPAGE)0;

  lpsztemplate = LoadResource(shell32_hInstance, hRes);
  if(lpsztemplate == NULL)
	  return (HPROPSHEETPAGE)0;

  memset(&ppage, 0x0, sizeof(PROPSHEETPAGE));
  ppage.dwSize = sizeof(PROPSHEETPAGE);
  ppage.dwFlags = PSP_DLGINDIRECT | PSP_USECALLBACK;
  ppage.pResource = lpsztemplate;
  ppage.pfnDlgProc = dlgproc;
  ppage.lParam = lParam;
  ppage.pfnCallback = pfnCallback;

  return CreatePropertySheetPageW(&ppage);
}

/*************************************************************************
 *
 * SH_FileTimerProc [Internal]
 *
 * is invoked every 100ms to check if the property sheet should be closed
 * required because the property sheet pages are MODELESS
 *
 */


VOID
CALLBACK 
SH_FileTimerProc(HWND hwnd, 
    UINT uMsg,
    UINT_PTR idEvent,
    DWORD dwTime)
{
  if(!SendMessage(hwnd, PSM_GETCURRENTPAGEHWND, (WPARAM)0, (LPARAM)0))
  {
	KillTimer(hwnd, (UINT_PTR)hwnd);
	DestroyWindow(hwnd);
  }
}

/*************************************************************************
 *
 * SH_FileGeneralFileType [Internal]
 *
 * retrieves file extension description from registry and sets it in dialog
 *
 * FIXME: retrieve file extension default icon and load it
 */

BOOL
SH_FileGeneralSetFileType(HWND hwndDlg, WCHAR * filext)
{
 WCHAR name[MAX_PATH];
 WCHAR value[MAX_PATH];
 DWORD lname = MAX_PATH;
 DWORD lvalue = MAX_PATH;

 HKEY hKey;
 LONG result;
 HWND hDlgCtrl = GetDlgItem(hwndDlg, 14005);

 if(filext == NULL || hDlgCtrl == NULL)
	 return FALSE;

 memset(name, 0x0, sizeof(TCHAR) * MAX_PATH);
 memset(value, 0x0, sizeof(TCHAR) * MAX_PATH);
 if(RegOpenKeyW(HKEY_CLASSES_ROOT, filext, &hKey) != ERROR_SUCCESS)
  return FALSE;

 result = RegEnumValueW(hKey,0, name, &lname, NULL, NULL, (LPBYTE)value, &lvalue);
 RegCloseKey(hKey);

 if(result != ERROR_SUCCESS)
   return FALSE;
 
 if(RegOpenKeyW(HKEY_CLASSES_ROOT, value, &hKey) == ERROR_SUCCESS)
 {
   lvalue = lname = MAX_PATH;
   memset(name, 0x0, sizeof(TCHAR) * MAX_PATH);
   memset(value, 0x0, sizeof(TCHAR) * MAX_PATH);
   result = RegEnumValueW(hKey,0, name, &lname, NULL, NULL, (LPBYTE)value, &lvalue);
   RegCloseKey(hKey);
 }
 
 if(result != ERROR_SUCCESS)
 {
   strcpyW(value, L"Unknown application"); //FIXME
 }
 /* file extension type */
 SendMessageW(hDlgCtrl, WM_SETTEXT, (WPARAM)NULL, (LPARAM)value);
 return TRUE;
}
/*************************************************************************
 *
 * SHFileGeneralGetFileTimeString [Internal]
 *
 * formats a given LPFILETIME struct into readable user format 
 *
 * FIXME: this function should look in the registry how to display the time
 */

BOOL 
SHFileGeneralGetFileTimeString(LPFILETIME lpFileTime, WCHAR * lpResult)
{
  FILETIME ft;
  SYSTEMTIME dt;
  WORD wYear;

  if(lpFileTime == NULL || lpResult == NULL)
    return FALSE;

  if (!FileTimeToLocalFileTime(lpFileTime, &ft))
	return FALSE;

  FileTimeToSystemTime(&ft, &dt);

  	wYear = dt.wYear;
    /* ddmmyy */
	sprintfW (lpResult, L"%02d/%02d/%04d  %02d:%02u",
					dt.wDay, dt.wMonth, wYear, dt.wHour, dt.wMinute);
	
	DPRINT("result %S \n",lpResult);
	return TRUE;
}

/*************************************************************************
 *
 * SH_FileGeneralSetText [Internal]
 *
 * sets file path string and filename string
 *
 */


BOOL 
SH_FileGeneralSetText(HWND hwndDlg, WCHAR * lpstr)
{
  int flength;
  int plength;
  WCHAR * lpdir;
  WCHAR buff[256];
  HWND hDlgCtrl;

  if(lpstr == NULL)
    return FALSE;

  lpdir = strrchrW(lpstr, L'\\'); /* find the last occurence of '\\' */

  plength = strlenW(lpstr);
  flength = strlenW(lpdir);
  if(lpdir)
  {
    /* location text field */
	strncpyW(buff, lpstr, plength - flength);
	buff[plength - flength] = UNICODE_NULL;
	hDlgCtrl = GetDlgItem(hwndDlg, 14009);
    SendMessageW(hDlgCtrl, WM_SETTEXT, (WPARAM)NULL, (LPARAM)buff);
  }

  if(flength > 1)
  {
    /* text filename field */
    strncpyW(buff, &lpdir[1], flength);
    hDlgCtrl = GetDlgItem(hwndDlg, 14001);
    SendMessageW(hDlgCtrl, WM_SETTEXT, (WPARAM)NULL, (LPARAM)buff);
  }

  return TRUE;
}

/*************************************************************************
 *
 * SH_FileGeneralSetFileTime [Internal]
 *
 * retrieves file information from file and sets in dialog
 */


BOOL
SH_FileGeneralSetFileTime(HWND hwndDlg, WCHAR * lpfilename)
{
  BOOL result;
  HANDLE hFile; 
  FILETIME create_time;
  FILETIME accessed_time;
  FILETIME write_time;
  WCHAR resultstr[60];
  HWND hDlgCtrl;

  if(lpfilename == NULL)
    return FALSE;
  
  hFile = CreateFileW(lpfilename,
                	GENERIC_READ,
					FILE_SHARE_READ,NULL,
                	OPEN_EXISTING,
		            FILE_ATTRIBUTE_NORMAL,
					NULL);
  if(hFile == INVALID_HANDLE_VALUE)
  {
    DPRINT("failed to open file \n");
    return FALSE;
  }
  result = GetFileTime(hFile, &create_time, &accessed_time, &write_time);
  CloseHandle(hFile);
  if(!result)
  {
    DPRINT("GetFileTime failed \n");
    return FALSE;
  }
  if(SHFileGeneralGetFileTimeString(&create_time,resultstr))
  {
    hDlgCtrl = GetDlgItem(hwndDlg, 14015);
	DPRINT("hDlgCtrl %x text %S \n",hDlgCtrl, resultstr);
	SendMessageW(hDlgCtrl, WM_SETTEXT, (WPARAM)NULL, (LPARAM)resultstr);
  }

  if(SHFileGeneralGetFileTimeString(&accessed_time, resultstr))
  {
    hDlgCtrl = GetDlgItem(hwndDlg, 14017);
	SendMessageW(hDlgCtrl, WM_SETTEXT, (WPARAM)NULL, (LPARAM)resultstr);
  }

  if(SHFileGeneralGetFileTimeString(&write_time, resultstr))
  {
    hDlgCtrl = GetDlgItem(hwndDlg, 14019);
	SendMessageW(hDlgCtrl, WM_SETTEXT, (WPARAM)NULL, (LPARAM)resultstr);
  }
  
  return TRUE;
}

/*************************************************************************
 *
 * SH_FileGeneralSetFileSize [Internal]
 *
 * retrieves file size and sets in dialog
 *
 */


BOOL
SH_FileGeneralSetFileSize(HWND hwndDlg, WCHAR * filename, PULARGE_INTEGER lpfilesize)
{
  HANDLE hFile;
  ULARGE_INTEGER result;
  DWORD dwFileHi;
  DWORD dwFileLo;
  WCHAR fileres[60];
  HWND hDlgCtrl;

  if(filename == NULL)
    return FALSE;

  hFile = CreateFileW(filename,
                	GENERIC_READ,
					FILE_SHARE_READ,NULL,
                	OPEN_EXISTING,
		            FILE_ATTRIBUTE_NORMAL,
					NULL);
  if(hFile == INVALID_HANDLE_VALUE)
  {
    DPRINT("failed to open file \n");
    return FALSE;
  }
 if(GetFileSize(hFile, NULL) == 0xFFFFFFFF)
 {
   DPRINT("GetFileSize failed \n");
   CloseHandle(hFile);
   return FALSE;
 }
 
 dwFileLo = GetFileSize(hFile, &dwFileHi);
 CloseHandle(hFile);
 
 if(dwFileLo == 0xFFFFFFFF && GetLastError() != NO_ERROR)
   return FALSE;
 
 result.LowPart = dwFileLo;
 result.HighPart = dwFileHi;

 /* FIXME result is casted from ULONGLONG to LONGLONG */ 
 if(!StrFormatByteSizeW((LONGLONG)result.QuadPart, fileres, sizeof(fileres)))
   return FALSE;

 hDlgCtrl = GetDlgItem(hwndDlg, 14011);
 DPRINT("result %S \n", fileres);
 SendMessageW(hDlgCtrl, WM_SETTEXT, (WPARAM)NULL, (LPARAM)fileres);
 if(lpfilesize)
 {
   lpfilesize->LowPart = dwFileLo;
   lpfilesize->HighPart = dwFileHi;
 }
 return TRUE;
}
/*************************************************************************
 *
 * SH_SetFileVersionText [Internal]
 *
 *
 */

BOOL
SH_FileVersionQuerySetText(HWND hwndDlg, DWORD dlgId, LPVOID pInfo, WCHAR * text, WCHAR ** resptr)
{
  UINT reslen;
  HWND hDlgCtrl;

  if(hwndDlg == NULL || resptr == NULL || text == NULL)
   return FALSE;

  if(VerQueryValueW(pInfo, text, (LPVOID *)resptr, &reslen))
  {
    /* file description property */
   hDlgCtrl = GetDlgItem(hwndDlg, dlgId);
   DPRINT("%S :: %S\n",text, *resptr);
   SendMessageW(hDlgCtrl, WM_SETTEXT, (WPARAM)0, (LPARAM)*resptr);
   return TRUE;
  }
  return FALSE;
}

/*************************************************************************
 *
 * SH_FileVersionQuerySetListText [Internal]
 *
 * retrieves a version string and adds it to listbox
 *
 */


BOOL 
SH_FileVersionQuerySetListText(HWND hwndDlg, LPVOID pInfo, WCHAR * text, WCHAR **resptr, WORD lang, WORD code)
{
  UINT reslen;
  HWND hDlgCtrl;
  UINT index;
  WCHAR buff[256];

  if(hwndDlg == NULL || resptr == NULL || text == NULL)
   return FALSE;

  sprintfW(buff, L"\\StringFileInfo\\%04x%04x\\%S", lang, code, text);

  if(VerQueryValueW(pInfo, buff, (LPVOID *)resptr, &reslen))
  {
    /* listbox name property */
   hDlgCtrl = GetDlgItem(hwndDlg, 14009);
   DPRINT("%S :: %S\n",text, *resptr);
   index = SendMessageW(hDlgCtrl, LB_ADDSTRING, (WPARAM)-1, (LPARAM)text);
   SendMessageW(hDlgCtrl, LB_SETITEMDATA, (WPARAM)index, (LPARAM)*resptr);
   return TRUE;
  }
  return FALSE;
}

/*************************************************************************
 *
 * SH_FileVersionInitialize [Internal]
 *
 * sets all file version properties in dialog
 */
BOOL
SH_FileVersionInitialize(HWND hwndDlg, WCHAR * lpfilename)
{
  LPVOID pBuf;
  DWORD versize;
  DWORD handle;
  LPVOID info = NULL;
  UINT infolen;
  WCHAR buff[256];
  HWND hDlgCtrl;
  WORD lang;
  WORD code;
  LPLANGANDCODEPAGE lplangcode;
  WCHAR * str;

  if(lpfilename == 0)
    return FALSE;

  if(!(versize = GetFileVersionInfoSizeW(lpfilename, &handle)))
  {
	DPRINT("GetFileVersionInfoSize failed \n");
    return FALSE;
  }
  
  if(!(pBuf = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, versize)))
  {
	DPRINT("HeapAlloc failed bytes %x \n",versize);
    return FALSE;
  }
  
  if(!GetFileVersionInfoW(lpfilename, handle, versize, pBuf))
  {
	HeapFree(GetProcessHeap(), 0, pBuf);
    return FALSE;
  }
  if(VerQueryValueW(pBuf, L"\\", &info, &infolen))
  {
    VS_FIXEDFILEINFO * inf = (VS_FIXEDFILEINFO *)info;
	sprintfW(buff, L"%d.%d.%d.%d",inf->dwFileVersionMS & 0xFFFF0000,
		                                  inf->dwFileVersionMS & 0x0000FFFF,
										  inf->dwFileVersionLS & 0xFFFF0000,
										  inf->dwFileVersionLS & 0x0000FFFF);

   hDlgCtrl = GetDlgItem(hwndDlg, 14001);
   DPRINT("MS %x LS %x res %S \n",inf->dwFileVersionMS, inf->dwFileVersionLS, buff);
   SendMessageW(hDlgCtrl, WM_SETTEXT, (WPARAM)NULL, (LPARAM)buff);
  }
  if(VerQueryValueW(pBuf, L"\\VarFileInfo\\Translation", (LPVOID *)&lplangcode, &infolen))
  {
    /* FIXME find language from current locale / if not available,
	 * default to english
	 * for now default to first available language
     */
    lang = lplangcode->lang;
    code = lplangcode->code;
  }
  
  sprintfW(buff, L"\\StringFileInfo\\%04x%04x\\FileDescription", lang, code);
  SH_FileVersionQuerySetText(hwndDlg, 14003, pBuf, buff, &str);
    
  sprintfW(buff, L"\\StringFileInfo\\%04x%04x\\LegalCopyright", lang, code);
  SH_FileVersionQuerySetText(hwndDlg, 14005, pBuf, buff, &str);

  /* listbox properties */
  SH_FileVersionQuerySetListText(hwndDlg, pBuf, L"CompanyName", &str, lang, code);
  SH_FileVersionQuerySetListText(hwndDlg, pBuf, L"FileVersion", &str, lang, code);
  SH_FileVersionQuerySetListText(hwndDlg, pBuf, L"InternalName", &str, lang, code);

  /* FIXME insert language identifier */

  SH_FileVersionQuerySetListText(hwndDlg, pBuf, L"OriginalFilename", &str, lang, code);
  SH_FileVersionQuerySetListText(hwndDlg, pBuf, L"ProductName", &str, lang, code);
  SH_FileVersionQuerySetListText(hwndDlg, pBuf, L"ProductVersion", &str, lang, code);
  SetWindowLong(hwndDlg, DWL_USER, (LONG)pBuf);
  return TRUE;
}

/*************************************************************************
 *
 * SH_FileVersionDlgProc
 *
 * wnd proc of 'Version' property sheet page
 */
INT_PTR 
CALLBACK 
SH_FileVersionDlgProc(   
    HWND hwndDlg,
    UINT uMsg,
    WPARAM wParam,
    LPARAM lParam
)
{
  LPPROPSHEETPAGE ppsp;
  WCHAR * lpstr;
  LPVOID * buf;
  switch(uMsg)
  {
  case WM_INITDIALOG:
	 ppsp = (LPPROPSHEETPAGE)lParam;
	 if(ppsp == NULL)
	   break;

	 DPRINT("WM_INITDIALOG hwnd %p lParam %p ppsplParam %x\n",hwndDlg, lParam, ppsp->lParam);

	 lpstr = (WCHAR *)ppsp->lParam;

	 if(lpstr == NULL)
	   break;

	 return SH_FileVersionInitialize(hwndDlg, lpstr);


  case WM_COMMAND:
     if(LOWORD(wParam) == 14009 && HIWORD(wParam) == LBN_DBLCLK)
	 {
       HWND hDlgCtrl;
	   LRESULT lresult;
	   WCHAR * str;
	   
	   hDlgCtrl = GetDlgItem(hwndDlg, 14009);
	   lresult = SendMessageW(hDlgCtrl, LB_GETCURSEL, (WPARAM)NULL, (LPARAM)NULL);
       if(lresult == LB_ERR)
	   {
	     break;
	   }
       str = (WCHAR *)SendMessageW(hDlgCtrl, LB_GETITEMDATA, (WPARAM)lresult, (LPARAM)NULL);

	   if(str == NULL)
	   {
		 break;
	   }
	   hDlgCtrl = GetDlgItem(hwndDlg, 14010);
       DPRINT("hDlgCtrl %x string %S \n",hDlgCtrl,str);
	   SendMessageW(hDlgCtrl, WM_SETTEXT, (WPARAM)NULL, (LPARAM)str);
       return TRUE;
	 }
  case WM_DESTROY:
       buf = (LPVOID)GetWindowLong(hwndDlg, DWL_USER);
	   HeapFree(GetProcessHeap(), 0, buf);
  default:
	  break;
  }
  return FALSE;
}

/*************************************************************************
 *
 * SH_FileGeneralDlgProc
 *
 * wnd proc of 'General' property sheet page
 *
 */

INT_PTR 
CALLBACK 
SH_FileGeneralDlgProc(   
    HWND hwndDlg,
    UINT uMsg,
    WPARAM wParam,
    LPARAM lParam
)
{
  LPPROPSHEETPAGE ppsp;
  WCHAR * lpstr;

  switch(uMsg)
  {
  case WM_INITDIALOG:
	 ppsp = (LPPROPSHEETPAGE)lParam;
	 if(ppsp == NULL)
	   break;

	 DPRINT("WM_INITDIALOG hwnd %p lParam %p ppsplParam %x\n",hwndDlg, lParam, ppsp->lParam);

	 lpstr = (WCHAR *)ppsp->lParam;

	 if(lpstr == NULL)
	   break;

    /* set general text properties filename filelocation and icon */
	SH_FileGeneralSetText(hwndDlg, lpstr);
	/* enumerate file extension from registry and application which opens it*/
    SH_FileGeneralSetFileType(hwndDlg, strrchrW(lpstr, '.'));
    /* set file time create/modfied/accessed */
	SH_FileGeneralSetFileTime(hwndDlg, lpstr);
    /* set file size */
	SH_FileGeneralSetFileSize(hwndDlg, lpstr, NULL);

	 return TRUE;
  default:
	  break;
  }
  return FALSE;
}

/*************************************************************************
 *
 * SH_ShowPropertiesDialog
 *
 * called from ShellExecuteExW32
 * 
 * lpf contains (quoted) path of folder/file
 *
 * FIXME: determine if path is an folder -> load not yet existant folder dialog resource 
 *        provide button change application type if file has registered type
 *        make filename field editable and apply changes to filename on close
 */

BOOL
SH_ShowPropertiesDialog(WCHAR * lpf)
{
  PROPSHEETHEADERW pinfo;
  HPROPSHEETPAGE hppages[MAX_PROPERTY_SHEET_PAGE];
  HPROPSHEETPAGE hpage;
  HWND hwnd;
  WCHAR wFileName[MAX_PATH];
  UINT numsheet = 0;
  DWORD fhandle = 0;

  DPRINT("SH_ShowPropertiesDialog entered\n");

  if(lpf== NULL)
	  return FALSE;
 
  memset(hppages, 0x0, sizeof(HPROPSHEETPAGE) * MAX_PROPERTY_SHEET_PAGE);

  if(lpf[0] == '"')
  {
	/* remove quotes from cmd */
     LPWSTR src = lpf + 1;
	 LPWSTR dst = wFileName;
	  
	 while(*src && *src!='"')
		*dst++ = *src++;

	    *dst = '\0';
  }
  else
    strcpyW(wFileName, lpf);

  hpage = SH_CreatePropertySheetPage("SHELL_FILE_GENERAL_DLG", SH_FileGeneralDlgProc, NULL, (LPARAM)wFileName);
  if(hpage == NULL)
    return FALSE;
  
  if(SH_InsertPropertySheetPage(hppages, hpage, MAX_PROPERTY_SHEET_PAGE))
    numsheet++;

  if(GetFileVersionInfoSizeW(wFileName, &fhandle))
  {
    hpage = SH_CreatePropertySheetPage("SHELL_FILE_VERSION_DLG", SH_FileVersionDlgProc, NULL, (LPARAM)wFileName);
    if(hpage != NULL)
    {
      if(SH_InsertPropertySheetPage(hppages, hpage, MAX_PROPERTY_SHEET_PAGE))
      numsheet++;
    }
  }
    
  memset(&pinfo, 0x0, sizeof(PROPSHEETHEADERW));
  pinfo.dwSize = sizeof(PROPSHEETHEADER);
  pinfo.dwFlags = PSH_MODELESS | PSH_NOCONTEXTHELP | PSH_PROPTITLE;
  pinfo.nPages = numsheet;
  pinfo.phpage = hppages;
  pinfo.pszCaption = wFileName;
  hwnd = (HWND)PropertySheetW(&pinfo);
  if(hwnd != INVALID_HANDLE_VALUE)
  {
    if(!SetTimer(hwnd, (UINT_PTR)hwnd, 100, SH_FileTimerProc))
	{
	  DestroyWindow(hwnd);
	  return FALSE;
	}
  }
  return TRUE;
}
/*EOF */
